import pytest


@pytest.fixture
def given_any_message():
    return "Hello world"


@pytest.fixture
def given_any_title():
    return "This is a title"


@pytest.fixture
def given_any_info_petisco():
    return {
        "app_name": "test",
        "app_version": "0.0.1",
        "petisco_version": "0.0.1",
        "environment": "Production",
    }


@pytest.fixture
def given_any_exception():
    return ValueError("Oops, something went wrong")


@pytest.fixture
def given_any_traceback():
    return """Traceback (most recent call last):
  File "e.py", line 7, in <module>
    raise TypeError("Again !?!")
TypeError: Again !?!
"""


@pytest.fixture
def given_long_traceback():
    return (
        """Traceback (most recent call last):
  File "e.py", line 7, in <module>
    raise TypeError("Again !?!")
TypeError: Again !?!
"""
        * 500
    )


@pytest.fixture
def given_any_input_parameters():
    return {"parameter_1": 1, "parameter_2": 2}



from typing import List, Union

from pika import BasicProperties
from pika.exceptions import ChannelClosedByBroker

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
    RabbitMqMessageConfigurer,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_queue_name_formatter import (
    RabbitMqMessageQueueNameFormatter,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqDomainEventBus(DomainEventBus):
    """
    An implementation of DomainEventBus using RabbitMQ infrastructure.
    Implementation is based on pika library.
    """

    def __init__(
        self,
        organization: str,
        service: str,
        connector: Union[
            RabbitMqConnector, RabbitMqConsumerConnector
        ] = RabbitMqConnector(),
    ):
        self.connector = connector
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"publisher-{self.exchange_name}"
        self.configurer = RabbitMqMessageConfigurer(organization, service, connector)
        self.properties = BasicProperties(delivery_mode=2)  # PERSISTENT_TEXT_PLAIN

    def publish(self, domain_event: DomainEvent) -> None:
        """
        Publish one DomainEvent
        """
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        domain_event = domain_event.update_meta(meta)
        try:
            channel = self.connector.get_channel(self.rabbitmq_key)
            routing_key = RabbitMqMessageQueueNameFormatter.format(
                domain_event, exchange_name=self.exchange_name
            )
            channel.confirm_delivery()
            channel.basic_publish(
                exchange=self.exchange_name,
                routing_key=routing_key,
                body=domain_event.json().encode(),
                properties=self.properties,
            )
            if channel.is_open and not isinstance(
                self.connector, RabbitMqConsumerConnector
            ):
                channel.close()

        except ChannelClosedByBroker:
            self._retry(domain_event)

    def publish_list(self, domain_events: List[DomainEvent]) -> None:
        """
        Publish a list of DomainEvent
        """
        meta = self.get_configured_meta()
        published_domain_event = []
        try:
            channel = self.connector.get_channel(self.rabbitmq_key)

            for domain_event in domain_events:
                self._check_is_domain_event(domain_event)
                domain_event = domain_event.update_meta(meta)
                routing_key = RabbitMqMessageQueueNameFormatter.format(
                    domain_event, exchange_name=self.exchange_name
                )
                channel.confirm_delivery()
                channel.basic_publish(
                    exchange=self.exchange_name,
                    routing_key=routing_key,
                    body=domain_event.json(),
                    properties=self.properties,
                )
                published_domain_event.append(domain_event)
            if channel.is_open and not isinstance(
                self.connector, RabbitMqConsumerConnector
            ):
                channel.close()
        except ChannelClosedByBroker:
            unpublished_domain_events = [
                event for event in domain_events if event not in published_domain_event
            ]
            self._retry_publish_list(unpublished_domain_events)

    def _retry(self, domain_event: DomainEvent) -> None:
        # If domain event queue is not configured, it will be configured and then try to publish again.
        self.configurer.configure()
        self.publish(domain_event)

    def _retry_publish_list(self, domain_events: List[DomainEvent]) -> None:
        # If domain event queue is not configured, it will be configured and then try to publish again.
        self.configurer.configure()
        self.publish_list(domain_events)

    def retry_publish_only_on_store_queue(self, domain_event: DomainEvent) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        domain_event = domain_event.update_meta(meta)

        channel = self.connector.get_channel(self.rabbitmq_key)
        channel.basic_publish(
            exchange=self.exchange_name,
            routing_key="retry.store",
            body=domain_event.json().encode(),
            properties=self.properties,
        )

    def retry_publish(
        self,
        domain_event: DomainEvent,
        retry_routing_key: str,
        retry_exchange_name: Union[str, None] = None,
    ) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        domain_event = domain_event.update_meta(meta)
        try:
            channel = self.connector.get_channel(self.rabbitmq_key)
            channel.confirm_delivery()

            retry_exchange = (
                retry_exchange_name
                if retry_exchange_name
                else f"retry.{self.exchange_name}"
            )

            channel.basic_publish(
                exchange=retry_exchange,
                routing_key=retry_routing_key,
                body=domain_event.json(),
                properties=self.properties,
            )
        except ChannelClosedByBroker:
            self._retry(domain_event)

    def close(self) -> None:
        """
        Close RabbitMQ connection.
        """
        self.connector.close(self.rabbitmq_key)



from typing import Any, Dict, List

from meiga import BoolResult, Error, Failure, Result, Success, isSuccess
from meiga.decorators import meiga

from petisco.base.application.patterns.crud_repository import (
    AggregateRootType,
    CrudRepository,
)
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
)
from petisco.base.domain.errors.defaults.not_found import AggregateNotFoundError
from petisco.base.domain.model.uuid import Uuid


class InmemoryCrudRepository(CrudRepository[AggregateRootType]):
    def __init__(self) -> None:
        self._data: Dict[Uuid, Any] = dict()

    @meiga
    def save(self, aggregate_root: AggregateRootType) -> BoolResult:
        if aggregate_root.aggregate_id in self._data:
            return Failure(AggregateAlreadyExistError(aggregate_root.aggregate_id))
        self._data[aggregate_root.aggregate_id] = aggregate_root
        return isSuccess

    @meiga
    def retrieve(self, aggregate_id: Uuid) -> Result[AggregateRootType, Error]:
        aggregate_root = self._data.get(aggregate_id)
        if aggregate_root is None:
            return Failure(AggregateNotFoundError(aggregate_id))
        return Success(aggregate_root)

    def update(self, aggregate_root: AggregateRootType) -> BoolResult:
        if aggregate_root.aggregate_id not in self._data:
            return Failure(AggregateNotFoundError(aggregate_root.aggregate_id))
        self._data[aggregate_root.aggregate_id] = aggregate_root
        return isSuccess

    def remove(self, aggregate_id: Uuid) -> BoolResult:
        if aggregate_id not in self._data:
            return Failure(AggregateNotFoundError(aggregate_id))
        self._data.pop(aggregate_id)
        return isSuccess

    def retrieve_all(self) -> Result[List[AggregateRootType], Error]:
        return Success(list(self._data.values()))



import os

ROOT_PATH = os.path.abspath(os.path.dirname(__file__))

with open(f"{ROOT_PATH}/VERSION") as f:
    __version__ = f.read().rstrip()

from petisco import public_api
from petisco.public_api import *

__all__ = public_api.__all__



from typing import Any, List, NoReturn, Type

from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class NotImplementedMessageConsumer(MessageConsumer[Any]):
    def start(self) -> NoReturn:
        pass

    def add_subscribers(self, subscribers: List[MessageSubscriber]) -> None:
        pass

    def add_subscriber_on_dead_letter(
        self, subscriber: Type[MessageSubscriber]
    ) -> None:
        pass

    def add_subscriber_on_queue(
        self,
        queue_name: str,
        subscriber: Type[MessageSubscriber],
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> None:
        pass

    def unsubscribe_subscriber_on_queue(self, queue_name: str) -> None:
        pass

    def resume_subscriber_on_queue(self, queue_name: str) -> None:
        pass

    def stop(self) -> None:
        pass



from typing import Optional

from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.errors.domain_error import DomainError
from petisco.extra.slack.application.notifier.blocks_slack_notifier_message_converter import (
    BlocksSlackNotifierMessageConverter,
    SlackNotifierMessageConverter,
)
from petisco.extra.slack.application.notifier.exception_blocks_slack_notifier_message_converter import (
    ExceptionBlocksSlackNotifierMessageConverter,
)


class SlackError(DomainError):
    pass


class SlackNotifier(Notifier):
    def __init__(
        self,
        token: str,
        channel: str,
        converter: Optional[
            SlackNotifierMessageConverter
        ] = BlocksSlackNotifierMessageConverter(),
        exception_converter: Optional[
            SlackNotifierMessageConverter
        ] = ExceptionBlocksSlackNotifierMessageConverter(),
    ):
        self.channel = channel
        self.converter = converter
        self.exception_converter = exception_converter
        self.client = WebClient(token=token)

    def publish(self, notifier_message: NotifierMessage) -> None:
        try:
            self.client.chat_postMessage(
                channel=self.channel,
                blocks=self.converter.convert(notifier_message),
                text=notifier_message.title,
            )
        except SlackApiError as e:
            raise SlackError(e.response["error"])

    def publish_exception(
        self, notifier_exception_message: NotifierExceptionMessage
    ) -> None:
        try:
            self.client.chat_postMessage(
                channel=self.channel,
                blocks=self.exception_converter.convert(notifier_exception_message),
                text=notifier_exception_message.title,
            )
        except SlackApiError as e:
            raise SlackError(e.response["error"])



import os
from distutils.util import strtobool
from typing import List, Optional

from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.dependency_injection.container import Container
from petisco.base.domain.message.message_subscriber import MessageSubscriber

MAX_RETRIES = 5

CLEAR_SUBSCRIBER_BEFORE = strtobool(
    os.getenv("PETISCO_RABBITMQ_CONFIGURER_CLEAR_SUBSCRIBER_BEFORE", "false")
)
CLEAR_STORE_BEFORE = strtobool(
    os.getenv("PETISCO_RABBITMQ_CONFIGURER_CLEAR_STORE_BEFORE", "false")
)


class RabbitMqMessageApplicationConfigurer(ApplicationConfigurer):
    def __init__(
        self,
        execute_after_dependencies: bool = True,
        subscribers: Optional[List[MessageSubscriber]] = None,
    ):
        self.execute_after_dependencies = execute_after_dependencies
        self.subscribers = subscribers
        super().__init__(execute_after_dependencies)

    def execute(self, testing: bool = False) -> None:
        configurer = Container.get("message_configurer")
        configurer.configure_subscribers(
            self.subscribers,
            clear_subscriber_before=CLEAR_SUBSCRIBER_BEFORE,
            clear_store_before=CLEAR_STORE_BEFORE,
        )



from typing import Type

from pydantic.main import BaseModel

from petisco.base.domain.message.message import Message


class MessageSubscriberInfo(BaseModel):
    message_name: str
    message_version: int
    message_type: str

    @staticmethod
    def from_class_type(class_type: Type[Message]) -> "MessageSubscriberInfo":
        message = class_type()
        return MessageSubscriberInfo(
            message_name=message.name,
            message_version=message.version,
            message_type=message.type,
        )



from collections import defaultdict
from typing import Any, Dict, List, Union

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.misc.singleton import Singleton


class Container(metaclass=Singleton):
    """
    Singleton which contains set dependencies (List[Dependency)) prepared to be instantiated in order to be injected
     in the UseCases of our application
    """

    def __init__(self) -> None:
        self.dependencies: Dict[str, Dependency] = defaultdict()

    @staticmethod
    def get(name: str) -> Any:
        """
        Returns an instance of set Dependency.
        """
        container = Container()
        dependency = container.dependencies.get(name)
        if dependency is None:
            raise IndexError(
                f"Invalid dependency. {name} is not found within available dependencies [{container.get_available_dependencies()}]"
            )
        instance = dependency.get_instance()
        return instance

    @staticmethod
    def set_dependencies(dependencies: Union[List[Dependency], None] = None) -> None:
        """
        Set dependencies from a list of them.
        """
        if dependencies is None:
            dependencies = []
        Container()._set_dependencies(dependencies)

    @staticmethod
    def get_available_dependencies() -> List[str]:
        """
        Returns the names (keys) of set dependencies.
        """
        return list(Container().dependencies.keys())

    def _set_dependencies(self, input_dependencies: List[Dependency]) -> None:
        for dependency in input_dependencies:
            if dependency.name not in self.dependencies:
                self.dependencies[dependency.name] = dependency



from typing import Union

from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.shared.queue_config import QueueConfig
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector
from petisco.extra.rabbitmq.shared.rabbitmq_declarer import RabbitMqDeclarer
from petisco.extra.rabbitmq.shared.rabbitmq_exchange_name_formatter import (
    RabbitMqExchangeNameFormatter,
)


class RabbitMqMessageStoreConfigurer:
    def __init__(
        self,
        organization: str,
        service: str,
        connector: Union[RabbitMqConnector, RabbitMqConsumerConnector],
        queue_config: QueueConfig,
    ) -> None:
        self._connector = connector
        self._exchange_name = f"{organization}.{service}"
        self._retry_exchange_name = RabbitMqExchangeNameFormatter.retry(
            self._exchange_name
        )
        self._dead_letter_exchange_name = RabbitMqExchangeNameFormatter.dead_letter(
            self._exchange_name
        )
        self._common_retry_exchange_name = f"retry.{organization}.store"
        self._common_dead_letter_exchange_name = f"dead_letter.{organization}.store"

        self.rabbitmq = RabbitMqDeclarer(
            connector=self._connector, channel_name=self._exchange_name
        )
        self.queue_config = queue_config

    def execute(self) -> None:
        self._configure_exchanges()
        self._declare_queues(
            self._exchange_name,
            self._retry_exchange_name,
            self._dead_letter_exchange_name,
        )

    def clear(self) -> None:
        self._delete_exchange()
        self._delete_queues()

    def _configure_exchanges(self) -> None:
        self.rabbitmq.declare_exchange(self._exchange_name)
        self.rabbitmq.declare_exchange(self._retry_exchange_name)
        self.rabbitmq.declare_exchange(self._dead_letter_exchange_name)
        self.rabbitmq.declare_exchange(self._common_retry_exchange_name)
        self.rabbitmq.declare_exchange(self._common_dead_letter_exchange_name)

    def _delete_exchange(self) -> None:
        self.rabbitmq.delete_exchange(self._exchange_name)
        self.rabbitmq.delete_exchange(self._retry_exchange_name)
        self.rabbitmq.delete_exchange(self._dead_letter_exchange_name)
        self.rabbitmq.delete_exchange(self._common_retry_exchange_name)
        self.rabbitmq.delete_exchange(self._common_dead_letter_exchange_name)

    def _delete_queues(self) -> None:
        self.rabbitmq.delete_queue("store")
        self.rabbitmq.delete_queue("retry.store")
        self.rabbitmq.delete_queue("dead_letter.store")

    def _declare_queues(
        self,
        exchange_name: str,
        retry_exchange_name: str,
        dead_letter_exchange_name: str,
    ) -> None:
        self.rabbitmq.declare_queue(
            queue_name="store",
            dead_letter_exchange=self._common_dead_letter_exchange_name,
            dead_letter_routing_key="dead_letter.store",
            message_ttl=self.queue_config.get_main_ttl("store"),
        )
        self.rabbitmq.declare_queue(
            queue_name="retry.store",
            dead_letter_exchange=self._common_retry_exchange_name,  # exchange_name
            dead_letter_routing_key="store",
            message_ttl=self.queue_config.get_retry_ttl("retry.store"),
        )
        self.rabbitmq.declare_queue(queue_name="dead_letter.store")

        for routing_key_any_message in ["*.*.*.event.*", "*.*.*.command.*"]:
            self.rabbitmq.bind_queue(
                exchange_name=exchange_name,
                queue_name="store",
                routing_key=routing_key_any_message,
            )
            self.rabbitmq.bind_queue(
                exchange_name=self._common_retry_exchange_name,
                queue_name="store",
                routing_key=routing_key_any_message,
            )
            self.rabbitmq.bind_queue(
                exchange_name=dead_letter_exchange_name,
                queue_name="dead_letter.store",
                routing_key=f"dead_letter.{routing_key_any_message}",
            )

        self.rabbitmq.bind_queue(
            exchange_name=exchange_name, queue_name="store", routing_key="retry.store"
        )
        self.rabbitmq.bind_queue(
            exchange_name=exchange_name, queue_name="store", routing_key="store"
        )

        self.rabbitmq.bind_queue(
            exchange_name=self._common_retry_exchange_name,
            queue_name="store",
            routing_key="store",
        )
        self.rabbitmq.bind_queue(
            exchange_name=self._common_retry_exchange_name,
            queue_name="retry.store",
            routing_key="retry.store",
        )
        self.rabbitmq.bind_queue(
            exchange_name=self._common_dead_letter_exchange_name,
            queue_name="dead_letter.store",
            routing_key="dead_letter.store",
        )
        self.rabbitmq.bind_queue(
            exchange_name=dead_letter_exchange_name,
            queue_name="dead_letter.store",
            routing_key="dead_letter.store",
        )



__all__ = []


from petisco.extra.slack.dependencies import get_default_notifier_dependencies
from petisco.extra.slack.is_slack_available import is_slack_available

__all__ = ["get_default_notifier_dependencies"]
if is_slack_available():
    from petisco.extra.slack.application.notifier.blocks_slack_notifier_message_converter import (
        BlocksSlackNotifierMessageConverter,
    )
    from petisco.extra.slack.application.notifier.exception_blocks_slack_notifier_message_converter import (
        ExceptionBlocksSlackNotifierMessageConverter,
    )
    from petisco.extra.slack.application.notifier.slack_notifier import SlackNotifier
    from petisco.extra.slack.application.notifier.slack_notifier_message_converter import (
        SlackNotifierMessageConverter,
    )
    from petisco.extra.slack.dependencies import get_slack_notifier_dependencies

    __all__ += [
        "SlackNotifier",
        "SlackNotifierMessageConverter",
        "BlocksSlackNotifierMessageConverter",
        "ExceptionBlocksSlackNotifierMessageConverter",
        "get_slack_notifier_dependencies",
    ]



from typing import Any, Generic, Type, TypeVar

T = TypeVar("T")


class Builder(Generic[T]):
    """
    Data structure that defines the creation of an instance from its arguments.
    """

    def __init__(
        self,
        klass: Type[T],
        name_constructor: str = None,
        is_builder: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.klass = klass
        self.args = args
        self.kwargs = kwargs
        self.name_constructor = name_constructor
        self.is_builder = is_builder

    def build(self) -> T:
        try:
            if self.name_constructor:
                constructor = getattr(self.klass, self.name_constructor)
                instance: T = constructor(*self.args, **self.kwargs)
            elif self.is_builder:
                instance: T = self.klass.build(*self.args, **self.kwargs)
            else:
                instance: T = self.klass(*self.args, **self.kwargs)
        except Exception as exc:
            raise RuntimeError(
                f"Error instantiating {self.klass.__name__}\n{repr(exc)}"
            )

        return instance



from abc import abstractmethod
from typing import Union

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message_bus import MessageBus


class DomainEventBus(MessageBus[DomainEvent]):
    """
    A base class to implement an infrastructure-based bus to publish events.
    """

    @abstractmethod
    def publish(self, domain_event: DomainEvent) -> None:
        """
        Publish one DomainEvent
        """
        raise NotImplementedError

    @abstractmethod
    def retry_publish_only_on_store_queue(self, domain_event: DomainEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def retry_publish(
        self,
        domain_event: DomainEvent,
        retry_routing_key: str,
        retry_exchange_name: Union[str, None] = None,
    ) -> None:
        raise NotImplementedError

    def _check_is_domain_event(self, domain_event: DomainEvent) -> None:
        if not domain_event or not issubclass(domain_event.__class__, DomainEvent):
            raise TypeError("DomainEventBus only publishes DomainEvent objects")

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError



from abc import ABC
from typing import Any, Dict


class Interface(ABC):
    """
    A base class to Interfaces
    """

    def info(self) -> Dict[str, Any]:
        return {"name": type(self).__name__}



import os
from functools import wraps
from inspect import signature
from typing import Any, Callable, Dict, List, Type, Union

import elasticapm
from meiga import Error, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.misc.result_mapper import ResultMapper


def get_middleware_instances(config: Dict[str, Any]) -> List[Middleware]:
    middlewares_instances: List[Middleware] = []

    middlewares_configs = getattr(config, "middlewares", [])
    if not middlewares_configs:
        middlewares_configs = get_middlewares_configuration_from_environment()

    for middlewares_config in middlewares_configs:
        if not isinstance(middlewares_config, Middleware):
            middlewares_instances.append(middlewares_config())
        else:
            middlewares_instances.append(middlewares_config)

    return middlewares_instances


def get_middlewares_configuration_from_environment():
    def gettype(name: str) -> Type[Middleware]:
        lookup_table = {
            "NotifierMiddleware": NotifierMiddleware,
            "PrintMiddleware": PrintMiddleware,
        }
        my_type = lookup_table.get(name)
        if not my_type:
            raise TypeError(
                f"Value {name} in PETISCO_DEFAULT_MIDDLEWARES is not valid. Please, use {list(lookup_table.keys())}"
            )
        return my_type

    default_middlewares_names: Union[str, None] = os.getenv(
        "PETISCO_DEFAULT_MIDDLEWARES"
    )
    if not default_middlewares_names:
        return []
    default_middlewares_names_list = default_middlewares_names.split(",")
    return [
        gettype(middleware_name) for middleware_name in default_middlewares_names_list
    ]


def wrapper(
    execute_func: Callable[..., Any],
    wrapped_class_name: str,
    config: Any,
    mapper: ResultMapper,
) -> Callable[..., Any]:
    @wraps(execute_func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        middlewares = get_middleware_instances(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")

        for middleware in middlewares:
            middleware.set_data(wrapped_class_name, arguments)
            middleware.before()

        try:
            result = execute_func(*args, **kwargs)
        except OnFailureException as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

        for middleware in middlewares:
            if result:
                middleware.after(result)

        mapped_result = mapper.map(result)

        return mapped_result

    return wrapped



from typing import Any, Dict, List, Optional

from pydantic.main import BaseModel


class NotifierMessage(BaseModel):
    title: str
    message: Optional[str] = None
    meta: Optional[Dict[str, Any]] = None
    files: Optional[List[Dict[str, Any]]] = None
    link: Optional[Dict[str, str]] = None
    # info_id: Optional[InfoId] = None



from abc import abstractmethod
from typing import Generic, List, NoReturn, Type, TypeVar

from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.base.misc.interface import Interface

T = TypeVar("T", bound=MessageSubscriber)


class MessageConsumer(Generic[T], Interface):
    """
    A base class to implement an infrastructure-based consumer to link received messages from rabbitmq with
    defined subscribers.
    """

    @classmethod
    def __repr__(cls) -> str:
        return cls.__name__

    @abstractmethod
    def add_subscribers(self, subscribers: List[T]) -> None:
        """
        Add defined subscribers to be connected with main queues.
        """
        raise NotImplementedError

    @abstractmethod
    def add_subscriber_on_dead_letter(self, subscriber: Type[T]) -> None:
        """
        Add defined subscribers to be connected with the correspondent dead letter.
        """
        raise NotImplementedError

    @abstractmethod
    def add_subscriber_on_queue(
        self,
        queue_name: str,
        subscriber: Type[T],
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> None:
        """
        Add defined subscribers to be connected with a specific queue name.
        """
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_subscriber_on_queue(self, queue_name: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def resume_subscriber_on_queue(self, queue_name: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def start(self) -> NoReturn:
        """
        Start the process of consuming messages from infrastructure and pass to subscriber.
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """
        Stop the process of consuming messages from infrastructure.
        """
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union


class Database(ABC):
    def __init__(self, name: str, models: Union[Dict[str, Any], None] = None):
        self.name = name
        self.models = models if models else {}

    def info(self) -> Dict[str, Any]:
        _info: Dict[str, Any] = {"name": self.name}
        if self.models:
            _info["models"] = self.models
        return _info

    @abstractmethod
    def create(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def clear_data(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_model(self, model_name: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def get_model_names(self) -> List[str]:
        raise NotImplementedError



from abc import abstractmethod
from typing import List

from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.base.misc.interface import Interface


class MessageConfigurer(Interface):
    @abstractmethod
    def configure_subscribers(
        self,
        subscribers: List[MessageSubscriber],
        clear_subscriber_before: bool = False,
        clear_store_before: bool = False,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError



from typing import Union

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus


class NotImplementedDomainEventBus(DomainEventBus):
    def publish(self, domain_event: DomainEvent) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def retry_publish_only_on_store_queue(self, domain_event: DomainEvent) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def retry_publish(
        self,
        domain_event: DomainEvent,
        retry_routing_key: str,
        retry_exchange_name: Union[str, None] = None,
    ) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def close(self) -> None:
        pass



import os

from setuptools import find_packages, setup

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_NAME = "petisco"
VERSION = open("petisco/VERSION").read().rstrip()

# The text of the README file
with open(os.path.join(CURRENT_DIR, "README.md")) as fid:
    README = fid.read()

with open("requirements/requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Petisco is a framework for helping Python developers to build clean Applications",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["DDD", "Use Case", "Clean Architecture", "REST", "Applications"],
    url="https://github.com/alice-biometrics/petisco",
    author="Alice Biometrics",
    author_email="support@alicebiometrics.com",
    license="MIT",
    install_requires=required,
    entry_points={
        "console_scripts": [
            "petisco = petisco.cli.petisco:main",
            "petisco-rabbitmq = petisco.cli.petisco_rabbitmq:main",
        ]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        # "Programming Language :: Python :: 3.11", # TODO
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    extras_require={
        "sqlalchemy": [
            "sqlalchemy==1.4.44",
            "sqlalchemy_utils==0.38.3",
            "PyMySQL==1.0.2",
        ],
        "redis": ["redis==4.3.4"],
        "rabbitmq": ["pika==1.3.1"],
        "slack": ["slack_sdk==3.19.3"],
        "elastic": ["elasticsearch<8.0.0,>=7.13.1", "elastic-apm==6.13.1"],
        "elastic-apm": ["elastic-apm==6.13.1"],
        "fastapi": ["fastapi==0.88.0"],
    },
)



from abc import ABC, abstractmethod
from typing import Any

from meiga import AnyResult


class Middleware(ABC):
    """
    A base class to model your Middleware.
    A Middleware works before some operations (Controller and Subscriber) and also before returning the final result.
    """

    wrapped_class_name: str = None
    wrapped_class_input_arguments: Any = None

    def set_data(
        self, wrapped_class_name: str, wrapped_class_input_arguments: Any
    ) -> None:
        self.wrapped_class_name = wrapped_class_name
        self.wrapped_class_input_arguments = wrapped_class_input_arguments

    @abstractmethod
    def before(self) -> None:
        pass

    @abstractmethod
    def after(self, result: AnyResult) -> None:
        pass



from typing import Any, Dict, Optional

from pydantic.main import BaseModel


class ConsumerDerivedAction(BaseModel):
    action: Optional[str] = None
    exchange_name: Optional[str] = None
    routing_key: Optional[str] = None
    headers: Optional[Dict[str, Any]] = None



from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, validator

DEFAULT_HTTP_ERROR_DETAIL = "Unknown Error"


class HttpError(BaseModel):
    status_code: Union[int, None] = 500
    detail: Optional[str] = DEFAULT_HTTP_ERROR_DETAIL
    headers: Optional[Dict[str, Any]] = None
    type_error: Optional[str] = None

    @validator("type_error", always=True)
    @classmethod
    def prevent_type_error_none(cls, value: Union[str, None]) -> str:
        if value is None:
            value = cls.__name__
        return value



import copy
from abc import abstractmethod
from typing import Any, Dict

from elasticapm.traces import Transaction, execution_context
from meiga import AnyResult, NotImplementedMethodError

from petisco import AppService


class ElasticApmMonitoringAppService(AppService):
    # This class is built on top of AppService and adds functionalities to instrument the services that are going to be
    # executed in different threads and that we want to keep monitoring.

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Dict[str, Any]) -> AnyResult:
        return NotImplementedMethodError

    def _set_transaction(self, transaction: Transaction) -> None:
        self.transaction = transaction

    def with_transaction(
        self, transaction: Transaction
    ) -> "ElasticApmMonitoringAppService":
        service = copy.copy(self)
        service._set_transaction(transaction)
        return service

    def monitoring(self) -> None:
        if hasattr(self, "transaction"):
            execution_context.set_transaction(self.transaction)



from typing import Any, Dict

from petisco.base.domain.errors.default_http_error_map import DEFAULT_HTTP_ERROR_MAP
from petisco.base.misc.result_mapper import ResultMapper
from petisco.extra.fastapi.controller.fastapi_failure_handler import (
    fastapi_failure_handler,
)
from petisco.extra.fastapi.controller.fastapi_success_handler import (
    fastapi_success_handler,
)


class FastAPIResultMapper:
    @staticmethod
    def default() -> ResultMapper:
        return ResultMapper(
            error_map=DEFAULT_HTTP_ERROR_MAP,
            success_handler=fastapi_success_handler,
            failure_handler=fastapi_failure_handler,
        )

    @staticmethod
    def from_config(config: Dict[str, Any]) -> ResultMapper:
        error_map = getattr(config, "error_map", DEFAULT_HTTP_ERROR_MAP)
        error_map = {**DEFAULT_HTTP_ERROR_MAP, **error_map}
        return ResultMapper(
            error_map=error_map,
            success_handler=getattr(config, "success_handler", fastapi_success_handler),
            failure_handler=fastapi_failure_handler,
        )



import os
import time
from typing import Dict

from loguru import logger
from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import StreamLostError

from petisco.base.misc.singleton import Singleton


class RabbitMqConnector(metaclass=Singleton):
    """
    Singleton class to define RabbitMQ connections and some infrastructure configurations.
    """

    def __init__(self) -> None:
        self.heartbeat = int(os.environ.get("RABBITMQ_HEARTBEAT", 60))
        self.user = os.environ.get("RABBITMQ_USER", "guest")
        self.password = os.environ.get("RABBITMQ_PASSWORD", "guest")
        self.host = os.environ.get("RABBITMQ_HOST", "localhost")
        self.port = os.environ.get("RABBITMQ_PORT", "5672")
        self.num_max_retries_connection = int(
            os.environ.get("RABBITMQ_CONNECTION_NUM_MAX_RETRIES", 15)
        )
        self.wait_seconds_retry = float(
            os.environ.get("RABBITMQ_CONNECTION_WAIT_SECONDS_RETRY", 1)
        )
        self.open_connections: Dict[str, BlockingConnection] = dict()

    def close_all(self) -> None:
        for key_connection in list(self.open_connections.keys()):
            connection = self.open_connections.pop(key_connection)
            if connection and connection.is_open:
                try:
                    connection.close()
                except StreamLostError as exc:
                    logger.warning(f"close_all: {str(exc)}")
                except Exception as exc:  # noqa
                    logger.error(f"close_all: {str(exc)}")

    def close(self, key_connection: str) -> None:
        connection = self.open_connections.pop(key_connection)
        if connection and connection.is_open:
            try:
                connection.close()
            except StreamLostError as exc:
                logger.warning(f"close: {str(exc)}")
            except Exception as exc:  # noqa
                logger.error(f"close_all: {str(exc)}")

    def get_connection(self, key_connection: str) -> BlockingConnection:
        connection = self.open_connections.get(key_connection)

        if not connection or not connection.is_open:
            connection = self._create_connection(key_connection)

        return connection

    def get_channel(self, key_connection: str) -> BlockingChannel:
        connection = self.get_connection(key_connection)
        try:
            channel = connection.channel()
        except StreamLostError:
            connection = self.get_connection(key_connection)
            channel = connection.channel()
        return channel

    def _create_connection(self, key_connection: str) -> BlockingConnection:

        connection = self._create_blocking_connection(key_connection)

        self._wait_for_open_connection(connection, key_connection)

        if not connection.is_open:
            time_elapsed = round(
                (self.wait_seconds_retry * self.num_max_retries_connection), 2
            )
            raise ConnectionError(
                f"RabbitMQConnector: Impossible to obtain a open connection with host {self.host}. "
                f"Retried {self.num_max_retries_connection} in ~{time_elapsed} seconds"
            )

        return connection

    def _create_blocking_connection(self, connection_name: str) -> BlockingConnection:
        try:
            connection = BlockingConnection(
                ConnectionParameters(
                    heartbeat=self.heartbeat,
                    host=self.host,
                    port=int(self.port),
                    credentials=PlainCredentials(
                        username=self.user, password=self.password
                    ),
                    client_properties={"connection_name": connection_name},
                )
            )
        except:  # noqa E722
            raise ConnectionError(
                f"RabbitMQConnector: Impossible to connect to host {self.host}. "
                f"Review the following envars: [RABBITMQ_USER, RABBITMQ_PASSWORD, RABBITMQ_HOST, RABBITMQ_PORT, RABBITMQ_CONNECTION_NUM_MAX_RETRIES, RABBITMQ_CONNECTION_WAIT_SECONDS_RETRY]"
            )
        return connection

    def _wait_for_open_connection(
        self, connection: BlockingConnection, key_connection: str
    ) -> None:
        for i in range(self.num_max_retries_connection):
            if connection.is_open:
                self.open_connections[key_connection] = connection
                break
            time.sleep(self.wait_seconds_retry)
